@@ -1,23 +1,23 @@
-from asyncio import all_tasks, current_task, ensure_future, gather, get_event_loop, sleep
+from asyncio import get_event_loop, sleep
 from contextlib import asynccontextmanager
-from dataclasses import dataclass
+from dataclasses import dataclass, field
 from datetime import datetime, timedelta
-from functools import cache, partial
+from functools import cache
 from importlib import import_module
-from signal import SIGKILL, SIGTERM
+from signal import SIGTERM
 from typing import (Any, AsyncGenerator, Generic, NamedTuple, Optional, Self,
                     TypeVar, get_type_hints)
 
 from sharded_queue.drivers import JsonTupleSerializer
 from sharded_queue.protocols import Lock, Serializer, Storage
-from sharded_queue.settings import settings
+from sharded_queue.settings import WorkerSettings
 
 T = TypeVar('T')
 
 
 class Route(NamedTuple):
-    thread: int = settings.default_thread
-    priority: int = settings.default_priority
+    thread: int = 0
+    priority: int = 0
 
 
 class Handler(Generic[T]):
@@ -36,10 +36,7 @@ class Handler(Generic[T]):
 
     @classmethod
     async def route(cls, *requests: T) -> list[Route]:
-        return [
-            Route(settings.default_thread, settings.default_priority)
-            for _ in requests
-        ]
+        return [Route() for _ in requests]
 
     async def start(self) -> None:
         pass
@@ -50,6 +47,9 @@ class Handler(Generic[T]):
     async def stop(self) -> None:
         pass
 
+    def batch_size(self) -> Optional[int]:
+        pass
+
 
 class Tube(NamedTuple):
     handler: type[Handler]
@@ -84,7 +84,9 @@ class Tube(NamedTuple):
 @dataclass
 class Queue(Generic[T]):
     def __init__(
-        self, storage: Storage, serializer: Optional[Serializer] = None
+        self,
+        storage: Storage,
+        serializer: Optional[Serializer] = None,
     ):
         self.storage = storage
         self.serializer = serializer or JsonTupleSerializer()
@@ -103,10 +105,23 @@ class Queue(Generic[T]):
 
         if recurrent:
             if_not_exists = True
+            tube = Tube(RecurrentHandler, Route())
+            length = await self.storage.length(tube.pipe)
+            messages = await self.storage.range(tube.pipe, length)
+
             pipe_messages = RecurrentHandler.transform(
                 pipe_messages, recurrent, self.serializer
             )
 
+            recurrent_tuples = [
+                (request.pipe, request.msg) for (_, request) in pipe_messages
+            ]
+
+            for msg in reversed(messages):
+                request = self.serializer.deserialize(RecurrentRequest, msg)
+                if (request.pipe, request.msg) in recurrent_tuples:
+                    await self.storage.remove(tube.pipe, msg)
+
         if defer:
             if_not_exists = True
             pipe_messages = DeferredHandler.transform(
@@ -132,12 +147,14 @@ class Queue(Generic[T]):
 class Worker:
     lock: Lock
     queue: Queue
+    pipe: Optional[str] = None
+    settings: WorkerSettings = field(default_factory=WorkerSettings)
 
     async def acquire_tube(
         self, handler: Optional[type[Handler]] = None
-    ) -> Tube:
+    ) -> Optional[Tube]:
         all_pipes = False
-        while True:
+        while get_event_loop().is_running():
             for pipe in await self.queue.storage.pipes():
                 if not await self.queue.storage.length(pipe):
                     continue
@@ -160,37 +177,34 @@ class Worker:
                 return tube
 
             if all_pipes:
-                await sleep(settings.worker_acquire_delay)
+                await sleep(self.settings.acquire_delay)
             else:
                 all_pipes = True
 
-    def page_size(self, limit: Optional[int] = None) -> int:
-        if limit is None:
-            return settings.worker_batch_size
-
-        return min(limit, settings.worker_batch_size)
+        return None
 
     async def loop(
         self,
         limit: Optional[int] = None,
         handler: Optional[type[Handler]] = None,
     ) -> None:
-        loop = get_event_loop()
+        get_event_loop().add_signal_handler(SIGTERM, self.housekeep)
         processed = 0
-        while True and limit is None or limit > processed:
+        while get_event_loop().is_running() and (
+            limit is None or limit > processed
+        ):
             tube = await self.acquire_tube(handler)
-            loop.add_signal_handler(SIGTERM, partial(self.stop, tube.pipe))
+            if not tube:
+                break
+            self.pipe = tube.pipe
             processed = processed + await self.process(tube, limit)
-            loop.remove_signal_handler(SIGTERM)
+            self.pipe = None
 
-    def stop(self, pipe: str) -> None:
-        get_event_loop().create_task(self.shutdown_worker(pipe))
+        get_event_loop().remove_signal_handler(SIGTERM)
 
-    async def shutdown_worker(self, pipe: str) -> None:
-        await self.lock.release(pipe)
-        tasks = [task for task in all_tasks() if task is not current_task()]
-        [task.cancel() for task in tasks]
-        await gather(*tasks)
+    def housekeep(self) -> None:
+        if self.pipe:
+            get_event_loop().create_task(self.lock.release(self.pipe))
 
     async def process(self, tube: Tube, limit: Optional[int] = None) -> int:
         deserialize = self.queue.serializer.deserialize
@@ -209,16 +223,22 @@ class Worker:
 
         async with tube.context() as instance:
             if isinstance(instance, DeferredHandler | RecurrentHandler):
-                instance.queue = self.queue
+                instance.worker = self
 
-            while limit is None or limit > processed_counter:
+            while get_event_loop().is_running() and (
+                limit is None or limit > processed_counter
+            ):
                 if tube.handler is RecurrentHandler:
-                    page_size = settings.recurrent_tasks_limit
+                    batch_size = self.settings.recurrent_tasks_limit
                 else:
-                    page_size = self.page_size(limit)
+                    batch_size = instance.batch_size()
+                    if batch_size is None:
+                        batch_size = self.settings.batch_size
+                    if limit is not None:
+                        batch_size = min(limit, batch_size)
                 processed = False
                 for pipe in pipes:
-                    msgs = await storage.range(pipe, page_size)
+                    msgs = await storage.range(pipe, batch_size)
                     if not len(msgs):
                         continue
 
@@ -227,9 +247,8 @@ class Worker:
                     ])
 
                     if tube.handler is RecurrentHandler:
-                        await self.lock.ttl(
-                            key=tube.pipe,
-                            ttl=settings.recurrent_check_interval
+                        await self.prolongate_lock(
+                            self.settings.recurrent_check_interval
                         )
                         return len(msgs)
 
@@ -242,23 +261,30 @@ class Worker:
 
                 if not processed:
                     empty_counter = empty_counter + 1
-                    if empty_counter >= settings.worker_empty_limit:
+                    if empty_counter >= self.settings.empty_limit:
                         break
-                    await sleep(settings.worker_empty_pause)
+                    await sleep(self.settings.empty_pause)
 
                 if tube.handler is DeferredHandler:
-                    await self.lock.ttl(
-                        key=tube.pipe,
-                        ttl=settings.deferred_retry_delay,
+                    await self.prolongate_lock(
+                        self.settings.deferred_retry_delay
                     )
                     return processed_counter
                 else:
-                    await self.lock.ttl(tube.pipe, settings.lock_timeout)
+                    await self.prolongate_lock()
 
-        await self.lock.release(tube.pipe)
+        if self.pipe:
+            await self.lock.release(self.pipe)
 
         return processed_counter
 
+    async def prolongate_lock(self, ttl: Optional[int] = None):
+        if not self.pipe:
+            raise RuntimeError('No active pipe')
+        if ttl is None:
+            ttl = self.lock.settings.timeout
+        await self.lock.ttl(self.pipe, ttl)
+
 
 class DeferredRequest(NamedTuple):
     timestamp: float
@@ -273,13 +299,13 @@ class DeferredRequest(NamedTuple):
 
         timestamp: float = now.timestamp()
         if not isinstance(delta, timedelta):
-            timestamp = delta = delta
+            timestamp = timestamp + delta
 
         return timestamp
 
 
 class DeferredHandler(Handler):
-    queue: Queue
+    worker: Worker
 
     async def handle(self, *requests: DeferredRequest) -> None:
         now: float = datetime.now().timestamp()
@@ -289,7 +315,8 @@ class DeferredHandler(Handler):
         ]
 
         if len(pending):
-            await self.queue.register(DeferredHandler, *pending)
+            pending = sorted(pending, key=lambda r: r.timestamp)
+            await self.worker.queue.register(DeferredHandler, *pending)
 
         todo: list[tuple[str, list]] = [
             (request.pipe, request.msg)
@@ -298,21 +325,22 @@ class DeferredHandler(Handler):
         ]
 
         for pipe in set([pipe for (pipe, _) in todo]):
-            ready = [
+            ready: list[str] = [
                 msg for msg in
                 [
-                    self.queue.serializer.serialize(request)
+                    self.worker.queue.serializer.serialize(request)
                     for (candidate_pipe, request) in todo
                     if candidate_pipe == pipe
                 ]
-                if not await self.queue.storage.contains(pipe, msg)
+                if not await self.worker.queue.storage.contains(pipe, msg)
             ]
 
+            ready = list(set(ready))
             if len(ready):
-                await self.queue.storage.append(pipe, *ready)
+                await self.worker.queue.storage.append(pipe, *ready)
 
         if len(pending) and not len(todo):
-            await sleep(settings.deferred_retry_delay)
+            await sleep(self.worker.settings.deferred_retry_delay)
 
     @classmethod
     def transform(
@@ -343,22 +371,22 @@ class RecurrentRequest(NamedTuple):
     @classmethod
     def get_interval(cls, interval: int | float | timedelta) -> float:
         if isinstance(interval, timedelta):
-            return float(interval.seconds)
+            return float(interval.total_seconds())
 
         return float(interval)
 
 
 class RecurrentHandler(Handler):
-    queue: Queue
+    worker: Worker
 
     async def handle(self, *requests: RecurrentRequest) -> None:
         deferred_pipe: str = Tube(DeferredHandler, Route()).pipe
         deferred_requests: list[tuple[str, str]] = [
             (request.pipe, request.msg)
             for request in [
-                self.queue.serializer.deserialize(DeferredRequest, msg)
-                for msg in await self.queue.storage.range(
-                    deferred_pipe, settings.recurrent_tasks_limit
+                self.worker.queue.serializer.deserialize(DeferredRequest, msg)
+                for msg in await self.worker.queue.storage.range(
+                    deferred_pipe, self.worker.settings.recurrent_tasks_limit
                 )
             ]
         ]
@@ -374,7 +402,7 @@ class RecurrentHandler(Handler):
         ]
 
         if len(todo):
-            await self.queue.register(
+            await self.worker.queue.register(
                 DeferredHandler, *todo, if_not_exists=True
             )
 
